########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations
from typing import Callable

from .. import astt
from ..error import Error, NoPrefixFunc, UnexpectedEOF, UnexpectedToken
from .. import token_type as ttoken

LOWEST = 1
EQUALS = 2
LESSGREATER = 3
SUM = 4
PRODUCT = 5
PREFIX = 6
CALL = 7
INDEX = 8


def parse(tokens: list[ttoken.Token], fname="stdin") -> tuple[astt.Program, list[Error]]:
    def advance(index: int) -> int:
        return index + 1 if index + 1 != len(tokens) else index

    def peek(index: int) -> ttoken.Token:
        return (
            tokens[index + 1] if tokens[index].ttype != ttoken.TT_EOF else tokens[index]
        )

    def get_precedence(token: ttoken.Token, precedences: dict[str, int]) -> int:
        if precedences.get(token.ttype) is not None:
            return precedences[token.ttype]
        return LOWEST

    def parse_statement(token: ttoken.Token, index: int) -> tuple[astt.LetStatement | astt.ReassignmentStatement | astt.ReturnStatement | astt.ExpressionStatement | None, int]:
        if token.ttype == ttoken.TT_LET:
            return parse_let(token, index, errors)
        elif token.ttype == ttoken.TT_RETURN:
            return parse_return(token, index)
        else:
            return parse_statementexpr(token, index)

    def parse_let(token: ttoken.Token, index: int, errors: list[Error]) -> tuple[astt.LetStatement | astt.ReassignmentStatement | None, int]:
        index = advance(index)

        if peek(index).ttype == ttoken.TT_ASSIGN:
            statement, index = parse_assign(token, index)
        elif peek(index).ttype == ttoken.TT_LBRACKET:
            statement, index = parse_reassign(token, index, errors)
        else:
            errors.append(UnexpectedToken(fname, "=", peek(index).string, (-1, -1)))
            return None, index

        return statement, index

    def parse_assign(token: ttoken.Token, index: int) -> tuple[astt.LetStatement | None, int]:
        statement = astt.LetStatement(token, astt.Identifier(token), astt.Expression(""))
        
        token = tokens[index]
        statement.name = astt.Identifier(token)
        
        index = advance(advance(index))
        token = tokens[index]
        value, index = parse_expression(LOWEST, token, index)

        if value is not None:
            statement.value = value
        
        if peek(index).ttype == ttoken.TT_SEMICOLON:
            index = advance(index)
            token = tokens[index]

        return statement, index

    def parse_reassign(token: ttoken.Token, index: int, errors: list[Error]) -> tuple[astt.ReassignmentStatement | None, int]:
        statement = astt.ReassignmentStatement(token, astt.IndexExpression(token, astt.ArrayLiteral(token, [astt.Expression("")]), astt.Expression("")), astt.Expression(""))
        
        token = tokens[index]
        index_expr, index = parse_expression(LOWEST, token, index)

        if index_expr is not None:
            assert isinstance(index_expr, astt.IndexExpression)
            statement.index_expr = index_expr
        
        if peek(index).ttype != ttoken.TT_ASSIGN:
            errors.append(UnexpectedToken(fname, "=", peek(index).string, (-1, -1)))
            return None, index

        index = advance(advance(index))
        token = tokens[index]
        value, index = parse_expression(LOWEST, token, index)

        if value is not None:
            statement.value = value
        
        if peek(index).ttype == ttoken.TT_SEMICOLON:
            index = advance(index)
            token = tokens[index]

        return statement, index

    def parse_return(token: ttoken.Token, index: int) -> tuple[astt.ReturnStatement, int]:
        statment = astt.ReturnStatement(token, astt.Expression(""))

        index = advance(index)
        token = tokens[index]

        value, index = parse_expression(LOWEST, token, index)
        if value is not None:
            statment.valuexp = value

        if peek(index).ttype == ttoken.TT_SEMICOLON:
            index = advance(index)
            token = tokens[index]

        return statment, index

    def parse_block(token: ttoken.Token, index: int) -> tuple[astt.BlockStatement, int]:
        block = astt.BlockStatement(token, [])

        index = advance(index)
        token = tokens[index]

        while token.ttype != ttoken.TT_RBRACE and token.ttype != ttoken.TT_EOF:
            statment, index = parse_statement(token, index)
            if statment is not None:
                block.statements.append(statment)
            index = advance(index)
            token = tokens[index]

        return block, index

    def parse_statementexpr(token: ttoken.Token, index: int) -> tuple[astt.ExpressionStatement, int]:
        statment = astt.ExpressionStatement(token, astt.Expression(""))
        expression, index = parse_expression(LOWEST, token, index)
        if expression is not None:
            statment.expression = expression

        if peek(index).ttype == ttoken.TT_SEMICOLON:
            index = advance(index)
            token = tokens[index]

        return statment, index

    def parse_expression(
        precedence: int, token: ttoken.Token, index: int
    ) -> tuple[astt.Expression | None, int]:
        parse_prefixfns: dict[str, Callable] = {
            ttoken.TT_IDEN: parse_identifier,
            ttoken.TT_NUM: parse_num,
            ttoken.TT_STRING: parse_str,
            ttoken.TT_BANG: parse_prefixexpr,
            ttoken.TT_MINUS: parse_prefixexpr,
            ttoken.TT_TRUE: parse_boolean,
            ttoken.TT_FALSE: parse_boolean,
            ttoken.TT_NULL: parse_null,
            ttoken.TT_LPAREN: parse_groupedexpr,
            ttoken.TT_IF: parse_ifexpr,
            ttoken.TT_WHILE: parse_while_expr,
            ttoken.TT_FOREACH: parse_foreach_expr,
            ttoken.TT_FUNC: parse_funcliteral,
            ttoken.TT_LBRACKET: parse_array_literal,
            ttoken.TT_LBRACE: parse_hash_literal,
        }
        precedences = {
            ttoken.TT_EQ: EQUALS,
            ttoken.TT_NTEQ: EQUALS,
            ttoken.TT_LT: LESSGREATER,
            ttoken.TT_GT: LESSGREATER,
            ttoken.TT_PLUS: SUM,
            ttoken.TT_MINUS: SUM,
            ttoken.TT_SLASH: PRODUCT,
            ttoken.TT_ASTERISK: PRODUCT,
            ttoken.TT_LPAREN: CALL,
            ttoken.TT_LBRACKET: INDEX,
        }

        parse_infixfns: dict[str, Callable] = {
            ttoken.TT_PLUS: parse_infixexpr,
            ttoken.TT_MINUS: parse_infixexpr,
            ttoken.TT_SLASH: parse_infixexpr,
            ttoken.TT_ASTERISK: parse_infixexpr,
            ttoken.TT_EQ: parse_infixexpr,
            ttoken.TT_NTEQ: parse_infixexpr,
            ttoken.TT_LT: parse_infixexpr,
            ttoken.TT_GT: parse_infixexpr,
            ttoken.TT_LPAREN: parse_callexpr,
            ttoken.TT_LBRACKET: parse_indexexpr,
        }

        prefix_func = parse_prefixfns.get(token.ttype)
        if prefix_func is None:
            errors.append(NoPrefixFunc(fname, token, (-1, -1)))
            return None, index
        left_expr, index = prefix_func(token, index)
        token = tokens[index]

        if left_expr is None:
            return None, index

        while peek(index) != ttoken.TT_SEMICOLON and precedence < get_precedence(
            tokens[advance(index)], precedences
        ):
            infix_func = parse_infixfns.get(peek(index).ttype)
            if infix_func is None:
                return left_expr

            index = advance(index)
            token = tokens[index]
            left_expr, index = infix_func(precedences, left_expr, token, index)

        return left_expr, index

    def parse_prefixexpr(token: ttoken.Token, index: int) -> tuple[astt.PrefixExpression | None, int]:
        expression = astt.PrefixExpression(token, astt.Expression(""))
        index = advance(index)
        token = tokens[index]
        if token.ttype == ttoken.TT_EOF:
            errors.append(UnexpectedEOF(fname, (-1, -1)))
            return None, index
        rightexpr, index = parse_expression(PREFIX, token, index)
        if rightexpr is None:
            return None, index

        expression.rightexpr = rightexpr

        return expression, index

    def parse_infixexpr(precedences: dict[str, int], leftexpr: astt.PrefixExpression, token: ttoken.Token, index: int) -> tuple[astt.InfixExpression | None, int]:
        expression = astt.InfixExpression(token, leftexpr, astt.Expression(""))
        precedence = get_precedence(token, precedences)

        index = advance(index)
        token = tokens[index]
        if token.ttype == ttoken.TT_EOF:
            errors.append(UnexpectedEOF(fname, (-1, -1)))
            return None, index
        rightexpr, index = parse_expression(precedence, token, index)
        if rightexpr is not None:
            expression.rightexpr = rightexpr

        return expression, index

    def parse_groupedexpr(token: ttoken.Token, index: int) -> tuple[astt.Expression | None, int]:
        index = advance(index)
        token = tokens[index]

        expression, index = parse_expression(LOWEST, token, index)

        if peek(index).ttype != ttoken.TT_RPAREN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_RPAREN, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(index)

        return expression, index

    def parse_ifexpr(token: ttoken.Token, index: int) -> tuple[astt.IfExpression | None, int]:
        expression = astt.IfExpression(
            token, astt.Expression(""), astt.BlockStatement(token, [astt.Statement("")]), astt.BlockStatement(token, [astt.Statement("")])
        )

        if peek(index).ttype != ttoken.TT_LPAREN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_LPAREN, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(advance(index))
        token = tokens[index]
        condition, index = parse_expression(LOWEST, token, index)
        if condition is not None:
            expression.condition = condition
        if peek(index).ttype != ttoken.TT_RPAREN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_RPAREN, peek(index).ttype, (-1, -1))
            )
            return None, index
        index = advance(index)
        token = tokens[index]
        if peek(index).ttype != ttoken.TT_LBRACE:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_LBRACE, peek(index).ttype, (-1, -1))
            )
            return None, index
        index = advance(index)
        token = tokens[index]
        consequence, index = parse_block(token, index)
        if condition is not None:
            expression.consequence = consequence
        if peek(index).ttype == ttoken.TT_ELSE:
            index = advance(index)
            token = tokens[index]
            if peek(index).ttype != ttoken.TT_LBRACE:
                errors.append(
                    UnexpectedToken(
                        fname, ttoken.TT_LBRACE, peek(index).ttype, (-1, -1)
                    )
                )
            index = advance(index)
            token = tokens[index]
            alternative, index = parse_block(token, index)
            if alternative is not None:
                expression.alternative = alternative

        return expression, index
    
    def parse_while_expr(token: ttoken.Token, index: int) -> tuple[astt.WhileExpression | None, int]:
        expression = astt.WhileExpression(token, astt.Expression(""), astt.BlockStatement(token, [astt.Statement("")]))

        if peek(index).ttype != ttoken.TT_LPAREN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_LPAREN, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(advance(index))
        token = tokens[index]
        condition, index = parse_expression(LOWEST, token, index)
        if condition is not None:
            expression.condition = condition
        if peek(index).ttype != ttoken.TT_RPAREN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_RPAREN, peek(index).ttype, (-1, -1))
            )
            return None, index
        index = advance(index)
        token = tokens[index]
        if peek(index).ttype != ttoken.TT_LBRACE:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_LBRACE, peek(index).ttype, (-1, -1))
            )
            return None, index
        index = advance(index)
        token = tokens[index]
        body, index = parse_block(token, index)
        if condition is not None:
            expression.body = body
 
        return expression, index

    def parse_foreach_expr(token: ttoken.Token, index: int) -> tuple[astt.ForEachExpression | None, int]:
        expression = astt.ForEachExpression(token, astt.Expression(""), astt.Identifier(token), astt.BlockStatement(token, [astt.Statement("")]))

        if peek(index).ttype != ttoken.TT_IDEN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_IDEN, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(index)
        token = tokens[index]

        expression.var = astt.Identifier(token)

        if peek(index).ttype != ttoken.TT_IN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_IN, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(advance(index))
        token = tokens[index]

        iterator, index = parse_expression(LOWEST, token, index)

        if iterator is not None:
            expression.iterator = iterator
        
        if peek(index).ttype != ttoken.TT_LBRACE:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_LBRACE, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(index)
        token = tokens[index] 

        body, index = parse_block(token, index)
        if body is not None:
            expression.body = body

        return expression, index

    def parse_funcliteral(token: ttoken.Token, index: int) -> tuple[astt.FunctionLiteral | None, int]:
        expression = astt.FunctionLiteral(token, [], astt.BlockStatement(token, [astt.Statement("")]))

        if peek(index).ttype != ttoken.TT_LPAREN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_LPAREN, peek(index).ttype, (-1, -1))
            )
            return None, index
        index = advance(index)
        token = tokens[index]
        identifiers, index = parse_parametres(index)
        if identifiers is not None:
            expression.parameters = identifiers

        if peek(index).ttype != ttoken.TT_LBRACE:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_LBRACE, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(index)
        token = tokens[index]

        body, index = parse_block(token, index)
        if body is not None:
            expression.body = body

        return expression, index

    def parse_parametres(index: int) -> tuple[list[astt.Identifier] | None, int]:
        identifiers = []

        if peek(index).ttype == ttoken.TT_RPAREN:
            index = advance(index)
            token = tokens[index]
            return identifiers, index

        index = advance(index)
        token = tokens[index]
        identifier = astt.Identifier(token)
        identifiers.append(identifier)

        while peek(index).ttype == ttoken.TT_COMMA:
            index = advance(advance(index))
            token = tokens[index]
            identifier = astt.Identifier(token)
            identifiers.append(identifier)

        if peek(index).ttype != ttoken.TT_RPAREN:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_RPAREN, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(index)
        token = tokens[index]

        return identifiers, index

    def parse_callexpr(precedences: dict[str, int], function: astt.FunctionLiteral, token: ttoken.Token, index: int) -> tuple[astt.CallExpression, int]:
        expression = astt.CallExpression(token, function, [])
        arguments, index = parse_exprlist(index, ttoken.TT_RPAREN)

        if arguments is not None:
            expression.arguments = arguments

        return expression, index

    def parse_hash_literal(token: ttoken.Token, index: int) -> tuple[astt.HashLiteral | None, int]:
        hash_literal = astt.HashLiteral(token, dict())

        while peek(index).ttype != ttoken.TT_RBRACE:
            index = advance(index)
            token = tokens[index]

            key, index = parse_expression(LOWEST, token, index)
            if peek(index).ttype != ttoken.TT_COLON:
                errors.append(
                    UnexpectedToken(fname, ttoken.TT_COLON, peek(index).ttype, (-1, -1))
                )
                return None, index
            index = advance(advance(index))
            if peek(index).ttype == ttoken.TT_EOF:
                errors.append(UnexpectedEOF(fname, (-1, -1)))
                return None, index
            token = tokens[index]
            value, index = parse_expression(LOWEST, token, index)
            if key is not None and value is not None:
                hash_literal.pairs[key] = value
            if (
                peek(index).ttype != ttoken.TT_RBRACE
                and peek(index).ttype != ttoken.TT_COMMA
            ):
                errors.append(
                    UnexpectedToken(fname, ttoken.TT_COMMA, peek(index).ttype, (-1, -1))
                )
                return None, index
            if peek(index).ttype == ttoken.TT_COMMA:
                index = advance(index)
                token = tokens[index]

        index = advance(index)

        return hash_literal, index

    def parse_array_literal(token: ttoken.Token, index: int) -> tuple[astt.ArrayLiteral, int]:
        array = astt.ArrayLiteral(token, [])

        exprlist, index = parse_exprlist(index, ttoken.TT_RBRACKET)
        if exprlist is not None:
            array.elements = exprlist

        return array, index

    def parse_exprlist(index: int, end: str) -> tuple[list[astt.Expression] | None, int]:
        exprlist = []

        if peek(index).ttype == end:
            index = advance(index)
            return exprlist, index

        index = advance(index)
        token = tokens[index]

        expression, index = parse_expression(LOWEST, token, index)

        if expression is not None:
            exprlist.append(expression)

        while peek(index).ttype == ttoken.TT_COMMA:
            index = advance(index)
            token = tokens[index]
            if token.ttype == ttoken.TT_EOF:
                errors.append(UnexpectedEOF(fname, (-1, -1)))
            index = advance(index)
            token = tokens[index]
            expression, index = parse_expression(LOWEST, token, index)
            token = tokens[index]
            if expression is not None:
                exprlist.append(expression)

        if peek(index).ttype != end:
            if peek(index).ttype == ttoken.TT_EOF:
                errors.append(UnexpectedEOF(fname, (-1, -1)))
            else:
                errors.append(UnexpectedToken(fname, end, peek(index).ttype, (-1, -1)))
            return None, index

        index = advance(index)

        return exprlist, index

    def parse_indexexpr(precedences: dict[str, int], left: astt.ArrayLiteral | astt.HashLiteral, token: ttoken.Token, index: int) -> tuple[astt.IndexExpression | None, int]:
        expression = astt.IndexExpression(token, left, astt.Expression(""))
        index = advance(index)
        token = tokens[index]
        if token == ttoken.TT_EOF:
            errors.append(UnexpectedEOF(fname, (-1, -1)))
            return None, index

        indexexpr, index = parse_expression(LOWEST, token, index)
        if indexexpr is not None:
            expression.index = indexexpr

        if peek(index).ttype != ttoken.TT_RBRACKET:
            errors.append(
                UnexpectedToken(fname, ttoken.TT_RBRACKET, peek(index).ttype, (-1, -1))
            )
            return None, index

        index = advance(index)

        return expression, index

    def parse_boolean(token: ttoken.Token, index: int) -> tuple[astt.Boolean, int]:
        return astt.Boolean(token, token.ttype == ttoken.TT_TRUE), index

    def parse_null(token: ttoken.Token, index: int) -> tuple[astt.Null, int]:
        return astt.Null(token), index

    def parse_num(token: ttoken.Token, index: int) -> tuple[astt.Number, int]:
        assert isinstance(token.value, float)
        return astt.Number(token), index

    def parse_str(token: ttoken.Token, index: int) -> tuple[astt.String, int]:
        return astt.String(token), index

    def parse_identifier(token: ttoken.Token, index: int) -> tuple[astt.Identifier, int]:
        return astt.Identifier(token), index

    program = astt.Program([])
    errors: list[Error] = []
    index = advance(-1)
    token = tokens[index]

    while token.ttype != ttoken.TT_EOF:
        if token.ttype == ttoken.TT_EOF:
            break
        statment, index = parse_statement(token, index)
        if statment is not None:
            program.statements.append(statment)
        index = advance(index)
        token = tokens[index]

    return program, errors



##############################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from __future__ import annotations
from sys import setrecursionlimit

from .. import astt
from ..objectt.enviroment import Enviroment
from .. import error
from ..evaluator.mbuiltins import builtins
from .. import objectt as obj

setrecursionlimit(10**6)


def evaluate(node: astt.Node, env: Enviroment, fname="stdin") -> obj.Object:
    def istruthy(cond_object: obj.Object):
        if cond_object == obj.NULL:
            return False
        elif cond_object == obj.TRUE:
            return True
        elif cond_object == obj.FALSE:
            return False
        else:
            return True

    def is_error(tobject: obj.Object) -> bool:
        return tobject.type() == obj.OBJ_ERROR

    def eval_program(statements: list[astt.Statement]) -> obj.Object:
        result = obj.Object()
        for statement in statements:
            result = evaluate(statement, env)

            if isinstance(result, obj.ReturnValue):
                return result.value
            elif isinstance(result, obj.Error):
                return result

        return result

    def eval_blockstmt(block: astt.BlockStatement) -> obj.Object:
        result = obj.Object()

        for statement in block.statements:
            result = evaluate(statement, env)

            if (
                result.type() == obj.OBJ_RETURN_VALUE
                or result.type() == obj.OBJ_ERROR
            ):
                return result

        return result

    def eval_prefixexpr(operator: str, right: obj.Object) -> obj.Object:
        if operator ==  "!":
            return eval_bang(right)
        elif operator ==  "-":
            return eval_minus(right)
        else:
            return obj.Error(
                error.UnknownOperator(fname, operator, None, right.type(), (-1, -1))
            )

    def eval_infixexpr(operator: str, left: obj.Object, right: obj.Object) -> obj.Object:
        if left.type() == obj.OBJ_NUM and right.type() == obj.OBJ_NUM:
            assert isinstance(left, obj.Number) and isinstance(right, obj.Number)
            return eval_num_infixexpr(operator, left, right)
        elif left.type() == obj.OBJ_STRING and right.type() == obj.OBJ_STRING:
            assert isinstance(left, obj.String) and isinstance(right, obj.String)
            return eval_str_infixexpr(operator, left, right)
        elif operator == "==":
            return obj.TRUE if left == right else obj.FALSE
        elif operator == "!=":
            return obj.TRUE if left != right else obj.FALSE
        else:
            return obj.Error(
                error.UnknownOperator(
                    fname, operator, left.type(), right.type(), (-1, -1)
                )
            )

    def eval_num_infixexpr(operator: str, left: obj.Number, right: obj.Number) -> obj.Object:
        leftval = left.value
        rightval = right.value

        if operator == "+":
            return obj.Number(leftval + rightval)
        elif operator == "-":
            return obj.Number(leftval - rightval)
        elif operator == "*":
            return obj.Number(leftval * rightval)
        elif operator == "/":
            if not rightval:
                return obj.Error(error.ZeroDivision(fname, leftval, (-1, -1)))
            return obj.Number(leftval / rightval)
        elif operator == "<":
            return obj.TRUE if leftval < rightval else obj.FALSE
        elif operator == ">":
            return obj.TRUE if leftval > rightval else obj.FALSE
        elif operator == "==":
            return obj.TRUE if leftval == rightval else obj.FALSE
        elif operator == "!=":
            return obj.TRUE if leftval != rightval else obj.FALSE
        else:
            return obj.Error(
                error.UnknownOperator(
                    fname, operator, left.type(), right.type(), (-1, -1)
                )
            )

    def eval_str_infixexpr(operator: str, left: obj.String, right: obj.String) -> obj.Object:
        if operator != "+":
            return obj.Error(
                error.UnknownOperator(
                    fname, operator, left.type(), right.type(), (-1, -1)
                )
            )
        left_val = left.value
        right_val = right.value

        return obj.String(left_val + right_val)

    def eval_exprs(expressions: list[astt.Expression], env: Enviroment) -> list[obj.Object]:
        result = []
        for expression in expressions:
            evaluated = evaluate(expression, env)
            if is_error(evaluated):
                return [evaluated]
            result.append(evaluated)

        return result

    def eval_ifexpr(ifelse: astt.IfExpression) -> obj.Object:
        condition = evaluate(ifelse.condition, env)
        if is_error(condition):
            return condition

        if istruthy(condition):
            return evaluate(ifelse.consequence, env)
        elif ifelse.alternative.literal != "if":
            return evaluate(ifelse.alternative, env)
        else:
            return obj.NULL

    def eval_while_expr(while_node: astt.WhileExpression) -> obj.Object:
        condition = evaluate(while_node.condition, env)
        if is_error(condition):
            return condition

        result = obj.NULL

        while istruthy(condition):
            result = evaluate(while_node.body, env)
            condition = evaluate(while_node.condition, env)

        return result

    def eval_foreach_expr(for_node: astt.ForEachExpression) -> obj.Object:
        iterator = evaluate(for_node.iterator, env)

        if not isinstance(iterator, obj.Iterable):
            return obj.Error("Not iterable")

        result = obj.NULL
        for element in iterator.get_iterable():
            env.set_iden(for_node.var.literal, element)
            result = evaluate(for_node.body, env)

        env.store.pop(for_node.var.literal, None)

        return result

    def eval_hashlit(node: astt.HashLiteral, env: Enviroment) -> obj.Object:
        pairs: dict[obj.HashKey, obj.HashPair] = dict()

        for key_node, val_node in node.pairs.items():
            key = evaluate(key_node, env)
            if is_error(key):
                return key
            if not isinstance(key, obj.Hashable):
                return obj.Error(error.UnsupporteKeyType(fname, key, (-1, -1)))
            value = evaluate(val_node, env)
            if is_error(value):
                return value
            hashed = key.hash_key()
            pairs[hashed] = obj.HashPair(key, value)

        return obj.Hash(pairs)

    def eval_indexexpr(left: obj.Object, indexexpr: obj.Object) -> obj.Object:
        if left.type() == obj.OBJ_ARRAY and indexexpr.type() == obj.OBJ_NUM:
            assert isinstance(left, obj.Array) and isinstance(indexexpr, obj.Number)
            return eval_arr_indexexpr(left, indexexpr)
        elif left.type() == obj.OBJ_STRING and indexexpr.type() == obj.OBJ_NUM:
            assert isinstance(left, obj.String) and isinstance(indexexpr, obj.Number)
            return eval_str_indexexpr(left, indexexpr)
        elif left.type() == obj.OBJ_HASH:
            assert isinstance(left, obj.Hash)
            return eval_hash_indexexpr(left, indexexpr)
        elif indexexpr.type() != obj.OBJ_NUM:
            return obj.Error(error.UnsupportedIndexType(fname, indexexpr, (-1, -1)))
        else:
            return obj.Error(error.UnsupportedIndexAccessType(fname, left, (-1, -1)))

    def eval_str_indexexpr(string: obj.String, indexexpr: obj.Number) -> obj.String:
        index = int(indexexpr.value)
        maximum = len(string.value) - 1

        if index < 0 or index > maximum:
            return obj.String("")
        return obj.String(string.value[index])

    def eval_hash_indexexpr(hash: obj.Hash, indexexpr: obj.Object) -> obj.Object:
        if not isinstance(indexexpr, obj.Hashable):
            return obj.Error(error.UnsupporteKeyType(fname, indexexpr, (-1, -1)))
        pair = hash.pairs.get(indexexpr.hash_key())
        if pair is None:
            return obj.NULL
        return pair.value

    def eval_hash_reassign(hash: obj.Hash, key: obj.Object, value: obj.Object) -> obj.Object:
        if not isinstance(key, obj.Hashable):
            return obj.Error(error.UnsupporteKeyType(fname, key, (-1, -1)))

        hash.pairs[key.hash_key()] = obj.HashPair(key, value)

        return value

    def eval_arr_indexexpr(array: obj.Array, indexexpr: obj.Number) -> obj.Object:
        index = int(indexexpr.value)
        maximum = len(array.elements) - 1

        if index < 0 or index > maximum:
            return obj.NULL
        return array.elements[index]

    def eval_arr_reassign(array: obj.Array, indexexpr: obj.Number, value: obj.Object) -> obj.Object:
        index = int(indexexpr.value)
        maximum = len(array.elements) - 1

        if index < 0 or index > maximum:
            return obj.Error("Out of Bounds")

        array.elements[index] = value

        return value

    def eval_identifier(node: astt.Identifier, env: Enviroment) -> obj.Object:
        value = env.get(node.literal)
        if value is not None:
            return value
        builtin = builtins.get(node.literal)
        if builtin is not None:
            return builtin

        return obj.Error(error.UnknownIdentifier(fname, node.literal, (-1, -1)))

    def eval_minus(right: obj.Object) -> obj.Object:
        if not isinstance(right, obj.Number):
            return obj.Error(
                error.UnknownOperator(fname, "-", None, right.type(), (-1, -1))
            )
        value = right.value
        return obj.Number(-value)

    def eval_bang(right: obj.Object) -> obj.Object:
        if right == obj.TRUE:
            return obj.FALSE
        elif right == obj.FALSE:
            return obj.TRUE
        elif right == obj.NULL:
            return obj.TRUE
        else:
            return obj.FALSE

    def apply_func(func: obj.Object, arguments: list[obj.Object]) -> obj.Object:
        if isinstance(func, obj.Function):
            if len(arguments) != len(func.parametres):
                return obj.Error("Mismatched number of args")
            extended_env = extend_funcenv(func, arguments)
            evaluated = evaluate(func.body, extended_env)
            return unwrap_rtrvalue(evaluated)
        elif isinstance(func, obj.Builtin):
            return func.func(fname, arguments)

        return obj.Error(error.NotAFunction(fname, func, (-1, -1)))

    def extend_funcenv(func: obj.Function, arguments: list[obj.Object]) -> Enviroment:
        env = Enviroment(func.env)
        for index, parameter in enumerate(func.parametres):
            env.set_iden(parameter.literal, arguments[index])
        return env

    def unwrap_rtrvalue(tobject: obj.Object) -> obj.Object:
        if isinstance(tobject, obj.ReturnValue):
            return tobject.value
        return tobject

    if isinstance(node, astt.Program):
        program: astt.Program = node
        return eval_program(program.statements)
    elif isinstance(node, astt.ExpressionStatement):
        return evaluate(node.expression, env)
    elif isinstance(node, astt.PrefixExpression):
        rightexpr = evaluate(node.rightexpr, env)
        if is_error(rightexpr):
            return rightexpr
        return eval_prefixexpr(node.operator, rightexpr)
    elif isinstance(node, astt.InfixExpression):
        left = evaluate(node.leftexpr, env)
        if is_error(left):
            return left
        right = evaluate(node.rightexpr, env)
        if is_error(right):
            return right
        return eval_infixexpr(node.operator, left, right)
    elif isinstance(node, astt.BlockStatement):
        return eval_blockstmt(node)
    elif isinstance(node, astt.FunctionLiteral):
        parametres = node.parameters
        body = node.body
        return obj.Function(parametres, body, env)
    elif isinstance(node, astt.CallExpression):
        function = evaluate(node.function, env)
        if is_error(function):
            return function
        arguments = eval_exprs(node.arguments, env)
        if len(arguments) == 1 and is_error(arguments[0]):
            return arguments[0]
        return apply_func(function, arguments)
    elif isinstance(node, astt.HashLiteral):
        return eval_hashlit(node, env)
    elif isinstance(node, astt.ArrayLiteral):
        elements = eval_exprs(node.elements, env)
        if len(elements) == 1 and is_error(elements[0]):
            return elements[0]
        return obj.Array(elements)
    elif isinstance(node, astt.IndexExpression):
        leftexpr = evaluate(node.left, env)
        if is_error(leftexpr):
            return leftexpr
        indexexpr = evaluate(node.index, env)
        if is_error(indexexpr):
            return indexexpr
        return eval_indexexpr(leftexpr, indexexpr)
    elif isinstance(node, astt.IfExpression):
        return eval_ifexpr(node)
    elif isinstance(node, astt.WhileExpression):
        return eval_while_expr(node)
    elif isinstance(node, astt.ForEachExpression):
        return eval_foreach_expr(node)
    elif isinstance(node, astt.ReturnStatement):
        value = evaluate(node.valuexp, env)
        if is_error(value):
            return value
        return obj.ReturnValue(value)
    elif isinstance(node, astt.LetStatement):
        value = evaluate(node.value, env)
        if is_error(value):
            return value
        env.set_iden(name=node.name.literal, value=value)
        return value
    elif isinstance(node, astt.ReassignmentStatement):
        structure = evaluate(node.index_expr.left, env)
        index = evaluate(node.index_expr.index, env)
        value = evaluate(node.value, env)
        if isinstance(structure, obj.Array):
            if not isinstance(index, obj.Number):
                return obj.Error("Type mismatch for index")
            return eval_arr_reassign(structure, index, value)
        elif isinstance(structure, obj.Hash):
            return eval_hash_reassign(structure, index, value)
        else:
            return obj.Error("Unsupported")

    elif isinstance(node, astt.Identifier):
        return eval_identifier(node, env)
    elif isinstance(node, astt.Number):
        return obj.Number(node.value)
    elif isinstance(node, astt.String):
        return obj.String(node.string)
    elif isinstance(node, astt.Boolean):
        if node.value:
            return obj.TRUE
        return obj.FALSE
    elif isinstance(node, astt.Null):
        return obj.NULL

    return obj.Error(error.UnknownNode(fname, (-1, -1)))



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations

from ..error import Error

TT_NUM = "TT_NUM"
TT_STRING = "TT_STRING"
TT_IDEN = "TT_IDEN"

TT_ASSIGN = "TT_ASSIGN"
TT_PLUS = "TT_PLUS"
TT_MINUS = "TT_MINUS"
TT_SLASH = "TT_SLASH"
TT_ASTERISK = "TT_ASTERISK"

TT_BANG = "TT_BANG"
TT_EQ = "TT_EQ"
TT_NTEQ = "TT_NTEQ"
TT_LT = "TT_LT"
TT_GT = "TT_GT"

TT_COMMA = "TT_COMMA"
TT_COLON = "TT_COLON"
TT_SEMICOLON = "TT_SEMICOLON"
TT_LBRACE = "TT_LBRACE"
TT_RBRACE = "TT_RBRACE"
TT_LPAREN = "TT_LPAREN"
TT_RPAREN = "TT_RPAREN"
TT_LBRACKET = "TT_LBRACKET"
TT_RBRACKET = "TT_RBRACKET"

TT_LET = "TT_LET"
TT_FUNC = "TT_FUNC"
TT_TRUE = "TT_TRUE"
TT_FALSE = "TT_FALSE"
TT_IF = "TT_IF"
TT_ELSE = "TT_ELSE"
TT_RETURN = "TT_RETURN"
TT_WHILE = "TT_WHILE"
TT_FOREACH = "TT_FOREACH"
TT_IN = "TT_IN"

TT_NULL = "TT_NULL"
TT_ERR = "TT_ERR"
TT_EOF = "TT_EOF"

keywords = {
    "fn": TT_FUNC,
    "let": TT_LET,
    "true": TT_TRUE,
    "false": TT_FALSE,
    "null": TT_NULL,
    "if": TT_IF,
    "else": TT_ELSE,
    "return": TT_RETURN,
    "while": TT_WHILE,
    "foreach": TT_FOREACH,
    "in": TT_IN
}

class Token:
    def __init__(self, ttype: str, string: str, value: Error | float | None) -> None:
        self.ttype = ttype
        self.string = string
        self.value = value
    def __repr__(self) -> str:
        if self.value is None and self.ttype != TT_IDEN:
            return f"{self.ttype}"
        elif self.ttype == TT_ERR:
            if isinstance(self.value, Error):
                return f"{self.ttype}: {self.value.errtype}"
        elif self.ttype == TT_IDEN:
            return f"{self.ttype}: {self.string}"
        return f"{self.ttype}: {self.string}"



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from ..token_type.token_type import (
    TT_ASSIGN as TT_ASSIGN,
    TT_ASTERISK as TT_ASTERISK,
    TT_BANG as TT_BANG,
    TT_COLON as TT_COLON,
    TT_COMMA as TT_COMMA,
    TT_ELSE as TT_ELSE,
    TT_EOF as TT_EOF,
    TT_EQ as TT_EQ,
    TT_ERR as TT_ERR,
    TT_FALSE as TT_FALSE,
    TT_FUNC as TT_FUNC,
    TT_GT as TT_GT,
    TT_IDEN as TT_IDEN,
    TT_IF as TT_IF,
    TT_WHILE as TT_WHILE,
    TT_FOREACH as TT_FOREACH,
    TT_IN as TT_IN,
    TT_LBRACE as TT_LBRACE,
    TT_LBRACKET as TT_LBRACKET,
    TT_LET as TT_LET,
    TT_LPAREN as TT_LPAREN,
    TT_LT as TT_LT,
    TT_MINUS as TT_MINUS,
    TT_NTEQ as TT_NTEQ,
    TT_NULL as TT_NULL,
    TT_NUM as TT_NUM,
    TT_PLUS as TT_PLUS,
    TT_RBRACE as TT_RBRACE,
    TT_RBRACKET as TT_RBRACKET,
    TT_RETURN as TT_RETURN,
    TT_RPAREN as TT_RPAREN,
    TT_SEMICOLON as TT_SEMICOLON,
    TT_SLASH as TT_SLASH,
    TT_STRING as TT_STRING,
    TT_TRUE as TT_TRUE,
    keywords as keywords,
    Token as Token,
)



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations
from .objectt import Object

class Enviroment:
    def __init__(self, outer: "Enviroment | None") -> None:
        self.store: dict[str, Object] = dict()
        self.outer = outer
    def get(self, name: str) -> Object | None:
        obj = self.store.get(name)
        if obj is None and self.outer is not None:
            obj = self.outer.get(name)
        return obj
    def set_iden(self, name: str, value: Object) -> Object:
        self.store[name] = value
        return value
    def __repr__(self) -> str:
        return repr(self.store)



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from .objectt import (
    OBJ_ARRAY as OBJ_ARRAY,
    OBJ_BOOLEAN as OBJ_BOOLEAN,
    OBJ_BUILTIN as OBJ_BUILTIN,
    OBJ_ERROR as OBJ_ERROR,
    OBJ_FUNCTION as OBJ_FUNCTION,
    OBJ_HASH as OBJ_HASH,
    OBJ_NULL as OBJ_NULL,
    OBJ_NUM as OBJ_NUM,
    OBJ_RETURN_VALUE as OBJ_RETURN_VALUE,
    OBJ_STRING as OBJ_STRING,
    Object as Object,
    HashKey as HashKey,
    Hashable as Hashable,
    Iterable as Iterable,
    HashPair as HashPair,
    Number as Number,
    String as String,
    Boolean as Boolean,
    ReturnValue as ReturnValue,
    Function as Function,
    Array as Array,
    Hash as Hash,
    Builtin as Builtin,
    Null as Null,
    Error as Error,
    TRUE as TRUE,
    FALSE as FALSE,
    NULL as NULL,
)
from .enviroment import Enviroment as Enviroment



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations 

from typing import TYPE_CHECKING, Callable

from .. import astt
from .. import error

if TYPE_CHECKING:
    from .enviroment import Enviroment

OBJ_NUM = "OBJ_NUM"
OBJ_STRING = "OBJ_STRING"
OBJ_BOOLEAN = "OBJ_BOOLEAN"
OBJ_NULL = "OBJ_NULL"
OBJ_RETURN_VALUE = "OBJ_RETURN_VALUE"
OBJ_FUNCTION = "OBJ_FUNCTION"
OBJ_ARRAY = "OBJ_ARRAY"
OBJ_HASH = "OBJ_HASH"
OBJ_BUILTIN = "OBJ_BUILTIN"
OBJ_ERROR = "OBJ_ERROR"

class Object:
    def type(self) -> str:
        return str()
    def inspect(self) -> str:
        return str()

class HashKey:
    def __init__(self, obj_type: str, value: float | str) -> None:
        self.obj_type = obj_type
        self.value = value
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, HashKey):
            return False
        return value.value == self.value
    def __ne__(self, value: object) -> bool:
        if not isinstance(value, HashKey):
            return True
        return not value.value == self.value
    def __hash__(self) -> int:
        return hash(f"{self.obj_type}: {self.value}")
    def __repr__(self) -> str:
        return f"{self.value}"

class Hashable:
    def hash_key(self) -> HashKey:
        return HashKey("", int())

class HashPair:
    def __init__(self, key: Object, value: Object) -> None:
        self.key = key
        self.value = value
    def __repr__(self) -> str:
        return f"{self.value}"

class Iterable:
    def get_iterable(self) -> list[Object]:
        return [Object()]

class Number(Hashable, Object):
    def __init__(self, value: float) -> None:
        self.value = value
    def type(self) -> str:
        return OBJ_NUM
    def inspect(self) -> str:
        return str(self.value)
    def hash_key(self) -> HashKey:
        return HashKey(self.type(), self.value)
    def __repr__(self) -> str:
        return self.inspect()

class String(Iterable, Hashable, Object):
    def __init__(self, value: str) -> None:
        self.value = value
    def get_iterable(self) -> list[Object]:
        return [String(char) for char in self.value]
    def type(self) -> str:
        return OBJ_STRING
    def inspect(self) -> str:
        return "\"" + self.value + "\""
    def hash_key(self) -> HashKey:
        return HashKey(self.type(), hash(self.value))
    def __repr__(self) -> str:
        return self.inspect()

class Boolean(Hashable, Object):
    def __init__(self, value: bool) -> None:
        self.value = value
    def type(self) -> str:
        return OBJ_BOOLEAN
    def inspect(self) -> str:
        return str(self.value).lower()
    def hash_key(self) -> HashKey:
        return HashKey(self.type(), 1 if self.value else 0)
    def __repr__(self) -> str:
        return self.inspect()

class ReturnValue(Object):
    def __init__(self, value: Object) -> None:
        self.value = value
    def type(self) -> str:
        return OBJ_RETURN_VALUE
    def inspect(self) -> str:
        return self.value.inspect()

class Function(Object):
    def __init__(self, parametres: list[astt.Identifier], body: astt.BlockStatement, env: "Enviroment") -> None:
        self.parametres = parametres
        self.body = body
        self.env = env
    def type(self) -> str:
        return OBJ_FUNCTION
    def inspect(self) -> str:
        return f"fn({', '.join([repr(parametre) for parametre in self.parametres])}) {repr(self.body)}"

class Array(Iterable, Object):
    def __init__(self, elements: list[Object]) -> None:
        self.elements = elements
    def get_iterable(self) -> list[Object]:
        return self.elements
    def type(self) -> str:
        return OBJ_ARRAY
    def inspect(self) -> str:
        return f"[{', '.join([element.inspect() for element in self.elements])}]"

class Hash(Iterable, Object):
    def __init__(self, pairs: dict[HashKey, HashPair]) -> None:
        self.pairs = pairs
    def get_iterable(self) -> list[Object]:
        return [Array([value.key, value.value]) for value in self.pairs.values()]
    def type(self) -> str:
        return OBJ_HASH
    def inspect(self) -> str:
        return f"{{{str([hashpair.key.inspect() + ': ' + hashpair.value.inspect() for hashpair in self.pairs.values()])[1:-1]}}}".replace("'", "")
    def __repr__(self) -> str:
        return self.inspect()

class Builtin(Object):
    def __init__(self, func: Callable) -> None:
        self.func = func
    def type(self) -> str:
        return OBJ_BUILTIN
    def inspect(self) -> str:
        return "Builtin Function"

class Null(Object):
    def type(self) -> str:
        return OBJ_NULL
    def inspect(self) -> str:
        return "null"
    def __repr__(self) -> str:
        return self.inspect()

class Error(Object):
    def __init__(self, error: str | error.Error) -> None:
        self.error = error
    def type(self) -> str:
        return OBJ_ERROR
    def inspect(self) -> str:
        return f"ERROR: {str(self.error)}"

TRUE = Boolean(True)
FALSE = Boolean(False)
NULL = Null()



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from .astt import (
    Node as Node,
    Program as Program,
    Statement as Statement,
    Expression as Expression,
    Identifier as Identifier,
    Number as Number,
    String as String,
    Boolean as Boolean,
    Null as Null,
    PrefixExpression as PrefixExpression,
    InfixExpression as InfixExpression,
    IfExpression  as IfExpression,
    WhileExpression as WhileExpression,
    ForEachExpression as ForEachExpression,
    FunctionLiteral as FunctionLiteral,
    ArrayLiteral as ArrayLiteral,
    HashLiteral as HashLiteral,
    LetStatement as LetStatement,
    ReassignmentStatement as ReassignmentStatement,
    ReturnStatement as ReturnStatement,
    CallExpression as CallExpression,
    IndexExpression as IndexExpression,
    BlockStatement as BlockStatement,
    ExpressionStatement as ExpressionStatement,
)



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations

from .. import error
from .. import objectt as obj


def arg_error(fname, expected, got, func):
    return obj.Error(error.ArgumentError(fname, expected, got, func, (-1, -1)))

def m_len(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return obj.Error(error.ArgumentError(fname, 1, len(args), "len", (-1, -1)))
    if type(args[0]) == obj.String:
        return obj.Number(len(args[0].value))
    elif type(args[0]) == obj.Array:
        return obj.Number(len(args[0].elements))
    else:
        return obj.Error(error.UnsupportedType(fname, args[0], "len", (-1, -1)))


def m_type(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return obj.Error(error.ArgumentError(fname, 1, len(args), "type", (-1, -1)))
    return obj.String(args[0].type())


def m_puts(_: str, args: list[obj.Object]) -> obj.Object:
    for arg in args:
        if type(arg) == obj.String:
            print(arg.inspect()[1:-1], end=" ")
        else:
            print(arg.inspect(), end=" ")
    print()

    return obj.NULL


def m_input(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return arg_error(fname, 1, len(args), "input")
    
    return obj.String(input(args[0].inspect().replace('"', "")))


def m_push(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 2:
        return obj.Error(error.ArgumentError(fname, 2, len(args), "push", (-1, -1)))
    if args[0].type() != obj.OBJ_ARRAY:
        return obj.Error(error.UnsupportedType(fname, args[0], "push", (-1, -1)))
    assert isinstance(args[0], obj.Array)
    arr = args[0].elements
    arr.append(args[1])

    return obj.NULL


def m_tail(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return arg_error(fname, 1, len(args), "tail")
    if args[0].type == obj.OBJ_ARRAY:
        assert isinstance(args[0], obj.Array)
        sequence = args[0].elements
        _, *tail = sequence
        return obj.Array(tail)
    elif args[0].type() == obj.OBJ_STRING:
        assert isinstance(args[0], obj.String)
        string = args[0].value
        tail = string[1:]
        return obj.String(tail)
    else:
        return obj.Error(error.UnsupportedType(fname, args[0], "tail", (-1, -1)))


def m_head(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return arg_error(fname, 1, len(args), "head")
    if args[0].type() == obj.OBJ_ARRAY:
        assert isinstance(args[0], obj.Array)
        arr = args[0].elements
        head, *_ = arr
        return obj.Array([head])
    elif args[0].type() == obj.OBJ_STRING:
        assert isinstance(args[0], obj.String)
        string = args[0].value
        head = string[0:-1]
        return obj.String(head)
    else:
        return obj.Error(error.UnsupportedType(fname, args[0], "head", (-1, -1)))


def m_first(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return arg_error(fname, 1, len(args), "first")
    if args[0].type() == obj.OBJ_ARRAY:
        assert isinstance(args[0], obj.Array)
        arr = args[0].elements
        start = arr[0]
        return start
    elif args[0].type() == obj.OBJ_STRING:
        assert isinstance(args[0], obj.String)
        string = args[0].value
        start = string[0]
        return obj.String(start)
    else:
        return obj.Error(error.UnsupportedType(fname, args[0], "first", (-1, -1)))


def m_last(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return arg_error(fname, 1, len(args), "last")
    if args[0].type() == obj.OBJ_ARRAY:
        assert isinstance(args[0], obj.Array)
        arr = args[0].elements
        end = arr[-1]
        return end
    elif args[0].type() == obj.OBJ_STRING:
        assert isinstance(args[0], obj.String)
        arr = args[0].value
        end = arr[-1]
        return obj.String(end)
    else:
        return obj.Error(error.UnsupportedType(fname, args[0], "last", (-1, -1)))


def m_insert(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 3:
        return arg_error(fname, 1, len(args), "insert")
    if args[0].type() != obj.OBJ_ARRAY:
       return obj.Error(error.UnsupportedType(fname, args[0], "insert", (-1, -1)))
    elif args[1].type() != obj.OBJ_NUM:
        return obj.Error(error.UnsupportedType(fname, args[0], "insert", (-1, -1)))

    assert isinstance(args[0], obj.Array)
    assert isinstance(args[1], obj.Number)
    arr = args[0].elements.copy()
    if len(arr) < args[1].value:
        arr = arr + [obj.NULL] * (int(args[1].value) - len(arr))
    arr.insert(int(args[1].value), args[2])

    return obj.Array(arr)


def m_change(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 3:
        return arg_error(fname, 1, len(args), "change")
    elif args[0].type() != obj.OBJ_ARRAY:
       return obj.Error(error.UnsupportedType(fname, args[0], "change", (-1, -1)))
    elif args[1].type() != obj.OBJ_NUM:
        return obj.Error(error.UnsupportedType(fname, args[0], "change", (-1, -1)))
    
    assert isinstance(args[0], obj.Array)
    assert isinstance(args[1], obj.Number)
    arr = args[0].elements.copy()
    if args[1].value >= len(arr):
        return obj.Error("Too high")
    arr[int(args[1].value)] = args[2]

    return obj.Array(arr)


def m_num(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return arg_error(fname, 1, len(args), "num")

    try:
        return obj.Number(float(args[0].inspect().replace('"', "")))
    except ValueError:
        return obj.Error("Value error, couldnt convert to number")

def m_str(fname: str, args: list[obj.Object]) -> obj.Object:
    if len(args) != 1:
        return arg_error(fname, 1, len(args), "str")
    
    return obj.String(args[0].inspect().replace('"', ""))


builtins: dict[str, obj.Builtin] = {
    "len": obj.Builtin(m_len),
    "type": obj.Builtin(m_type),
    "puts": obj.Builtin(m_puts),
    "input": obj.Builtin(m_input),
    "push": obj.Builtin(m_push),
    "tail": obj.Builtin(m_tail),
    "head": obj.Builtin(m_head),
    "first": obj.Builtin(m_first),
    "last": obj.Builtin(m_last),
    "change": obj.Builtin(m_change),
    "insert": obj.Builtin(m_insert),
    "num": obj.Builtin(m_num), 
    "str": obj.Builtin(m_str)
}
